# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from rich import print_json
from pathlib import Path
from typer import Argument, Option, Typer
from typing import List, Optional, Tuple
from typing_extensions import Annotated

from ..api import Acceleration, AccessMode, Predictor, PredictorStatus, PredictorType
from .auth import get_access_key
from .misc import create_learn_callback

app = Typer(no_args_is_help=True)

@app.command(name="retrieve", help="Retrieve a predictor.")
def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    predictor = Predictor.retrieve(tag, access_key=get_access_key())
    predictor = asdict(predictor) if predictor else None
    print_json(data=predictor)

@app.command(name="search", help="Search predictors.")
def search_predictors (
    query: str=Argument(..., help="Search query."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    predictors = Predictor.search(query=query, offset=offset, count=count, access_key=get_access_key())
    predictors = [asdict(predictor) for predictor in predictors]
    print_json(data=predictors)

@app.command(name="create", help="Create a predictor.")
def create_predictor (
    tag: str=Argument(..., help="Predictor tag."),
    type: PredictorType=Argument(..., case_sensitive=False, help="Predictor type."),
    notebook: Path=Argument(..., help="Path to predictor notebook."),
    access: AccessMode=Option(None, case_sensitive=False, help="Predictor access mode."),
    description: str=Option(None, help="Predictor description."),
    media: Path=Option(None, help="Predictor image."),
    acceleration: Acceleration=Option(None, case_sensitive=False, help="Predictor acceleration."),
    environment: Annotated[Optional[List[str]], Option(default=[], help="Predictor environment variables.")] = None,
    license: str=Option(None, help="Predictor license URL."),
    overwrite: Annotated[bool, Option("--overwrite")] = None
):
    environment = { e.split("=")[0].strip(): e.split("=")[1].strip() for e in environment }
    predictor = Predictor.create(
        tag,
        type,
        notebook,
        access=access,
        description=description,
        media=media,
        acceleration=acceleration,
        environment=environment,
        license=license,
        overwrite=overwrite
    )
    predictor = asdict(predictor)
    print_json(data=predictor)

@app.command(name="delete", help="Delete a predictor.")
def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    result = Predictor.delete(tag, access_key=get_access_key())
    print_json(data=result)

@app.command(name="archive", help="Archive an active predictor.")
def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    predictor = Predictor.archive(tag, access_key=get_access_key())
    print_json(data=asdict(predictor))

@app.callback()
def predictor_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.fxn.ai/predictors"), help="Learn about predictors in Function.")
):
    pass


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from base64 import b64encode
from dataclasses import dataclass
from filetype import guess_mime
from io import BytesIO
from numpy import ndarray
from pathlib import Path
from PIL import Image
from typing import Any, Dict, List, Optional, Union

from .dtype import Dtype

@dataclass(frozen=True)
class FeatureInput:
    """
    Prediction input feature.

    Members:
        name (str): Feature name. This MUST match the input parameter name defined by the predictor.
        data (str): Feature data URL. This can be a web URL or a data URL.
        type (Dtype): Feature data type.
        shape (list): Feature shape. This MUST be provided for array features.
    """
    name: str
    data: str = None
    type: Dtype = None
    shape: Optional[List[int]] = None
    stringValue: str = None
    floatValue: float = None
    floatArray: List[float] = None
    intValue: int = None
    intArray: List[int] = None
    boolValue: bool = None
    listValue: list = None
    dictValue: Dict[str, Any] = None

    @classmethod
    def from_value (
        cls,
        value: Union[ndarray, str, float, int, bool, List, Dict[str, any], Path, Image.Image],
        name: str
    ) -> FeatureInput:
        """
        Create a feature input from a given value.

        Parameters:
            value (any): Value.
            name (str): Feature name.

        Returns:
            FeatureInput: Feature input.
        """
        # Array
        if isinstance(value, ndarray):
            encoded_data = b64encode(value).decode("ascii")
            data = f"data:application/octet-stream;base64,{encoded_data}"
            return FeatureInput(name, data, value.dtype.name, list(value.shape))
        # String
        if isinstance(value, str):
            return FeatureInput(name, stringValue=value)
        # Float
        if isinstance(value, float):
            return FeatureInput(name, floatValue=value)
        # Boolean
        if isinstance(value, bool):
            return FeatureInput(name, boolValue=value)
        # Integer
        if isinstance(value, int):
            return FeatureInput(name, intValue=value)
        # List
        if isinstance(value, list):
            return FeatureInput(name, listValue=value)
        # Dict
        if isinstance(value, dict):
            return FeatureInput(name, dictValue=value)
        # Image
        if isinstance(value, Image.Image):
            image_buffer = BytesIO()
            channels = { "L": 1, "RGB": 3, "RGBA": 4 }[value.mode]
            format = "PNG" if value.mode == "RGBA" else "JPEG"
            value.save(image_buffer, format=format)
            encoded_data = b64encode(image_buffer.getvalue()).decode("ascii")
            data = f"data:{value.get_format_mimetype()};base64,{encoded_data}"
            shape = [1, value.height, value.width, channels]
            return FeatureInput(name, data, Dtype.image, shape)
        # Path
        if isinstance(value, Path):
            assert value.is_file(), "Input path must be a file, not a directory"
            value = value.expanduser().resolve()
            type = _file_to_dtype(value)
            mime = guess_mime(str(value)) or "application/octet-stream"
            with open(value, "rb") as f:
                buffer = BytesIO(f.read())
            encoded_data = b64encode(buffer.getvalue()).decode("ascii")
            data = f"data:{mime};base64,{encoded_data}"
            return FeatureInput(name, data, type)
        # Unsupported
        raise RuntimeError(f"Cannot create input feature for value {value} of type {type(value)}")
    
def _file_to_dtype (path: Path) -> str:
    mime = guess_mime(str(path))
    if not mime:
        return Dtype.binary
    if mime.startswith("image"):
        return Dtype.image
    if mime.startswith("video"):
        return Dtype.video
    if mime.startswith("audio"):
        return Dtype.audio
    if path.suffix in [".obj", ".gltf", ".glb", ".fbx", ".usd", ".usdz", ".blend"]:
        return Dtype._3d
    return Dtype.binary


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from typer import Argument, Context, Option, Typer

def predict ( # INCOMPLETE
    tag: str = Argument(..., help="Predictor tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Generate raw output features instead of parsing."),
    context: Context = 0
):
    print("Predict!")


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from .dtype import Dtype
from .feature import Feature
from .featureinput import FeatureInput
#from .prediction import Prediction # INCOMPLETE
from .predictor import Acceleration, AccessMode, Parameter, Predictor, PredictorStatus, PredictorType, Signature
from .profile import Profile
from .storage import Storage, UploadType
from .user import User